(* Franois Thomasset -- INRIA Rocquencourt -- Octobre 2001 *)

(* Reprsentation intermdiaire de Maple *)

(*
Copyright  2001-2002 Franois Thomasset, all rights reserved.
Copying is covered by the GNU General Public License (GPL).
 
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
 
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.
 
$Revision: 1.1.1.1 $
$Date: 2002/03/08 08:31:57 $

*)

type name_string =
    Ident of string
  | QuotedString of string ;;
type oper =
    PlusOp | MinusOp | MultOp | DivOp
  | ExpoOp
  | OrOp | AndOp | LTOp | GTOp | LEOp | GEOp | NEOp | EQOp
  | ModOp
  | Union | Minus | Intersect
  | NeutralMulOp
  | NeutralOp of string ;;
type unary_op = UnaryMinus | UnaryPlus | NotOp ;;
type sepstmt = SemiColon | Colon | NoSep ;;
type name =
    NameString of name_string
  | NameIndexed of name * expr list
  | NameConcat of name * name_concat
and name_concat =
    N_int of int
  | N_str of string
  | N_expr of expr
and expr =
  | ArrowExpr of name list * expr
  | BinExpr of bin_expr
  | Factorial of expr
  | UnaryExpr of unary_op * expr
  | SeqExpr of expr option * expr
  | IntervalExpr of expr * expr
  | NameExpr of name
  | StringExpr of string
  | IntExpr of int
  | FloatExpr of int * int
  | CallExpr of call_expr
  | FunctionalCallExpr of functional_call_expr
  | UnevaluatedExpr of expr
  | ListExpr of expr list
  | SetExpr of expr list
  | ProcDef of proc_def
  | ParenExpr of expr list
  | ComposeExpr of expr * expr
  | RepeatComposeExpr of expr * expr
and bin_expr = { left: expr; bin_op: oper; right: expr }
and call_expr = { callee: name; args: expr list }
and functional_call_expr = { callee_f: name; fargs: (expr list) list }
and proc_def = { params: param list;
		 result_type: name_string option;
		 locals: name list;
		 options: name list;
		 body_of_proc: stat_comment_sep list
	       }
and param = { param_name: name; param_type: name_string option }

and stat =
    AssignStat of assign_stat
  | ExprStat of expr
  | ReadStat of expr
  | SaveStat of name list * name_string
  | IfThenElseStat of ite_stat
  | ForStat of for_stat
  | ForInStat of for_in_stat
  | StopStat
  | EmptyStat

and assign_stat = { lhs_of_assign: name list; rhs: expr list }

and ite_stat = { cond: expr;
		 then_part: stat_comment_sep list;
		 alternatives: alt list;
		 else_part: stat_comment_sep list
	       }
and alt = { cond_of_alt: expr; stats_of_alt: stat_comment_sep list }

and for_stat = { loop_index : name option ;
		 from_expr : expr option ;
		 by_expr : expr option ;
		 to_expr : expr option ;
		 while_expr : expr option ;
		 body : stat_comment_sep list
	       } 

and for_in_stat = { loop_var : name;
		    in_expr : expr;
		    in_while : expr option;
		    in_body : stat_comment_sep list
		  }

(* Le commentaire se place devant le statement *)
and stat_comment_sep = { comments: string list; sep: sepstmt; stat: stat } ;;

type prog = stat_comment_sep list ;;

(* Utilitaires pour debug *)
(* Rendus obsoletes par les fonctions dans kel.ml *)
let string_of_expr = function
  | ArrowExpr (_,_) -> "ArrowExpr"
  | BinExpr (_) -> "BinExpr"
  | Factorial (_) -> "Factorial"
  | UnaryExpr (_,_) -> "UnaryExpr"
  | SeqExpr (_,_) -> "SeqExpr"
  | IntervalExpr (_,_) -> "IntervalExpr"
  | NameExpr (_) -> "NameExpr"
  | StringExpr (_) -> "StringExpr"
  | IntExpr (_) -> "IntExpr"
  | FloatExpr (_) -> "FloatExpr"
  | CallExpr (_) -> "CallExpr"
  | FunctionalCallExpr (_) -> "CallExpr"
  | UnevaluatedExpr (_) -> "UnevaluatedExpr"
  | ListExpr (_) -> "ListExpr"
  | SetExpr (_) -> "SetExpr"
  | ProcDef (_) -> "ProcDef"
  | ParenExpr (_) -> "ParenExpr"
  | ComposeExpr (_,_) -> "ComposeExpr"
  | RepeatComposeExpr (_,_) -> "RepeatComposeExpr" ;;
let string_of_stat = function
    AssignStat (_) -> "AssignStat"
  | ExprStat (_) -> "ExprStat"
  | ReadStat (_) -> "ReadStat"
  | SaveStat (_) -> "SaveStat"
  | IfThenElseStat (_) -> "IfThenElseStat"
  | ForStat (_) -> "ForStat"
  | ForInStat (_) -> "ForInStat"
  | StopStat -> "StopStat"
  | EmptyStat -> "EmptyStat" ;;
