(* Franois Thomasset -- INRIA Rocquencourt -- Octobre 2001 *)
(*
   $Revision: 1.1.1.1 $
   $Date: 2002/03/08 08:31:58 $
*)

type name_string = Ident of string | QuotedString of string
and oper =
    PlusOp
  | MinusOp
  | MultOp
  | DivOp
  | ExpoOp
  | OrOp
  | AndOp
  | LTOp
  | GTOp
  | LEOp
  | GEOp
  | NEOp
  | EQOp
  | ModOp
  | Union
  | Minus
  | Intersect
  | NeutralMulOp
  | NeutralOp of string
and unary_op = UnaryMinus | UnaryPlus | NotOp
and sepstmt = SemiColon | Colon | NoSep
and name =
    NameString of name_string
  | NameIndexed of name * expr list
  | NameConcat of name * name_concat
and name_concat = N_int of int | N_str of string | N_expr of expr
and expr =
    ArrowExpr of name list * expr
  | BinExpr of bin_expr
  | Factorial of expr
  | UnaryExpr of unary_op * expr
  | SeqExpr of expr option * expr
  | IntervalExpr of expr * expr
  | NameExpr of name
  | StringExpr of string
  | IntExpr of int
  | FloatExpr of int * int
  | CallExpr of call_expr
  | FunctionalCallExpr of functional_call_expr
  | UnevaluatedExpr of expr
  | ListExpr of expr list
  | SetExpr of expr list
  | ProcDef of proc_def
  | ParenExpr of expr list
  | ComposeExpr of expr * expr
  | RepeatComposeExpr of expr * expr
and bin_expr = { left : expr; bin_op : oper; right : expr; }
and call_expr = { callee : name; args : expr list; }
and functional_call_expr = { callee_f : name; fargs : (expr list) list; }
and proc_def = {
  params : param list;
  result_type : name_string option;
  locals : name list;
  options : name list;
  body_of_proc : stat_comment_sep list;
} 
and param = { param_name : name; param_type : name_string option }
and stat =
    AssignStat of assign_stat
  | ExprStat of expr
  | ReadStat of expr
  | SaveStat of name list * name_string
  | IfThenElseStat of ite_stat
  | ForStat of for_stat
  | ForInStat of for_in_stat
  | StopStat
  | EmptyStat
and assign_stat = { lhs_of_assign : name list; rhs : expr list; } 
and ite_stat = {
  cond : expr;
  then_part : stat_comment_sep list;
  alternatives : alt list;
  else_part : stat_comment_sep list;
} 
and alt = { cond_of_alt : expr; stats_of_alt : stat_comment_sep list; } 
and for_stat = {
  loop_index : name option;
  from_expr : expr option;
  by_expr : expr option;
  to_expr : expr option;
  while_expr : expr option;
  body : stat_comment_sep list;
} 
and for_in_stat = {
  loop_var : name;
  in_expr : expr;
  in_while : expr option;
  in_body : stat_comment_sep list;
} 
and stat_comment_sep = {
  comments : string list;
  sep : sepstmt;
  stat : stat;
} 
and prog = stat_comment_sep list
val string_of_expr : expr -> string
val string_of_stat : stat -> string
