(* Franois Thomasset -- INRIA Rocquencourt -- Octobre 2001 *)

(* Translation from Maple to MuPad : main program *)

(*
Copyright  2001-2002 Franois Thomasset, all rights reserved.
Copying is covered by the GNU General Public License (GPL).
 
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
 
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.
 
$Revision: 1.3 $
$Date: 2002/03/08 08:48:38 $
*)
exception PARSE_ERROR;;
let version = "2.0 - 8 Mars 2002 - $Revision: 1.3 $";;
let verbose = ref false;;
let evaluation_level = ref (Some 100)
and list_of_names = ref []
and ind = ref 1
in
let printed_version = ref false
in
let print_version () =
  if !printed_version
  then ()
  else
    begin
      printed_version := true;
      Printf.printf "Maple to MuPad - %s - Franois Thomasset -- INRIA Rocquencourt\n" version
    end
in
let print_usage () =
  print_version();
  Printf.printf "Usage: maple_to_mupad <MapleFile>\n";
  Printf.printf "Or   : maple_to_mupad -l <level> <MapleFile>\n";
  Printf.printf "Or   : maple_to_mupad -nl <MapleFile>\n"
in
let read_argument () =
  match Sys.argv.(!ind) with
  | "-nl" -> evaluation_level := None
  | "-l" ->
      begin
	if !ind >= (Array.length (Sys.argv) - 1)
	then
	  begin
	    print_usage(); flush stdout; exit(-1)
	  end
	else
	  begin
	    ind := !ind + 1;
	    try
	      evaluation_level := Some (int_of_string (Sys.argv.(!ind)))
	    with
	      Failure _ ->
		begin
		  Printf.printf "Error at position %d in command line:\n" !ind;
		  Printf.printf "Option -l must be followed by a number";
		  print_usage();
		  flush stdout;
		  exit(-1)
		end
	  end
      end
  | "-V" -> print_version()
  | "-v" -> verbose := true
  | "-h" -> print_usage()
  | name ->
      begin
	if (Sys.argv.(!ind).[0] = '-')
	then
	  begin
	    Printf.printf "Unknown option: %s\n" Sys.argv.(!ind);
	    print_usage()
	  end
	else
	  list_of_names := (!list_of_names @ [name])
      end
in
while (!ind < Array.length (Sys.argv)) do
  read_argument ();
  ind := !ind + 1
done;
let translate_a_file file_name =
  let _ = begin if (!verbose) then Printf.printf "Translating %s\n" file_name end
  in
  let file_suffixe =
    if Filename.check_suffix file_name "mpl"
    then
      Filename.chop_suffix file_name "mpl"
    else (file_name ^ ".")
  in
  let in_ch = open_in file_name in
  let lexbuf = Lexing.from_channel in_ch in
  let result =
    try Maple_parser.program Maple_lexer.token lexbuf
    with
      Parsing.Parse_error ->
        begin
          print_newline();
	  Printf.printf "Parse error in line %d\n" (1 + !Maple_lexer.linecount) ;
          Printf.printf "Position : %d %d\n" (Lexing.lexeme_start lexbuf) (Lexing.lexeme_end lexbuf);
          close_in in_ch;
          flush stdout;
	  raise PARSE_ERROR
        end
  in
  begin
    close_in in_ch;
    let ch_prt_mupad = open_out (file_suffixe ^ "mupad") in
    (* insrer une affectation  LEVEL au dbut de chaque fichier *)
    Mpl2mup.prt_prog ch_prt_mupad (!evaluation_level) (!verbose) result ;
    flush ch_prt_mupad;
    Parsing.clear_parser ();
    if (!verbose) then Printf.printf "Translated %s into %s: OK\n" file_name (file_suffixe ^ "mupad")
  end
in
let rec translate_all_files = function
  | [] -> ()
  | n :: r ->
      begin
	try
	  translate_a_file n
	with
	  PARSE_ERROR -> ();
	translate_all_files r
      end
in
if (!verbose) then print_version();
translate_all_files (!list_of_names);;
