{
(* Franois Thomasset -- INRIA Rocquencourt -- Octobre 2001 *)

(* Translation from Maple to MuPad : lexical specification of maple *)

(*
Copyright  2001-2002 Franois Thomasset, all rights reserved.
Copying is covered by the GNU General Public License (GPL).
 
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
 
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.
 
$Revision: 1.1.1.1 $
$Date: 2002/03/08 08:31:57 $

*)

 open Lexing;;
 open Maple_parser;;
 let linecount = ref 0;;
}

rule token = parse
  [' ' '\t' '\r']     { token lexbuf }
| '\n'			{ incr linecount; token lexbuf }
| '#' [ ^ '\n' ]* '\n'	{ incr linecount;
			  COMMENT ( String.sub (lexeme lexbuf) 1 ( String.length (lexeme lexbuf) - 2 ) ) }
| ';'			{ SEMICOLON }
| ':'			{ COLON }
| "::"			{ DOUBLE_COLON }
| ":="			{ ASSIGN }
| '+'			{ PLUS }
| '-'			{ SUBTRACT }
| '*'			{ MULT }
| "**"			{ EXP }
| '/'			{ SLASH }
| '^'			{ CARET }
| '!'			{ EXCLAM }
| '='			{ EQUAL }
| '<'			{ LT }
| '>'			{ GT }
| "<="			{ LE }
| ">="			{ GE }
| "<>"			{ NE }
| "not"			{ NOT }
| "and"			{ AND }
| "or"			{ OR }
| "->"			{ ARROW }
| '@'			{ AT }
| "@@"			{ REPEAT_COMPOSE }
| "&*"			{ AMPMUL }
| '&' [ ^' ' '\t' '\r' '\n' '*' '&' '(' ')' '[' ']' '{' '}' ';' ':' '\'' '`' '#' ]+
			{ AMPOP ( String.sub (lexeme lexbuf) 1 ( String.length (lexeme lexbuf) - 1 ) ) }
| '$'			{ SEQ }
| ".."			{ ELLIPSE }
| '.'			{ DOT }
| ','			{ COMMA }
| '('			{ LPAREN }
| ')'			{ RPAREN }
| '['			{ LBRACK }
| ']'			{ RBRACK }
| '{'			{ LBRACE }
| '}'			{ RBRACE }
| '\''			{ QUOTE }
| '|'			{ BAR }
| '_'			{ UNDERSCORE }
| '%'			{ PERCENT }
| '\\'			{ BACKSLASH }
| '?'			{ QUESTIONMARK }
| "by"			{ BY }
| "do"			{ DO }
| "done"		{ DONE }
| "elif"		{ ELIF }
| "else"		{ ELSE }
| "end"			{ END }
| "fi"			{ FI }
| "for"			{ FOR }
| "from"		{ FROM }
| "if"			{ IF }
| "in"			{ IN }
| "global"		{ GLOBAL }
| "local"		{ LOCAL }
| "od"			{ OD }
| "option"		{ OPTION }
| "proc"		{ PROC }
| "quit"		{ QUIT }
| "read"		{ READ }
| "save"		{ SAVE }
| "stop"		{ STOP }
| "then"		{ THEN }
| "to"			{ TO }
| "while"		{ WHILE }
| "union"		{ UNION }
| "intersect"		{ INTERSECT }
| "minus"		{ MINUS }
| "mod"			{ MOD }
| ['0'-'9']+		{ INT ( int_of_string ( Lexing.lexeme lexbuf ) ) }
| ['a'-'z''A'-'Z''_'] ['0'-'9''a'-'z''A'-'Z''_']*
			{ ID(Lexing.lexeme lexbuf) }
| '`' [ ^ '\n' '`' ]+ '`' 
			{ QUOTED_STRING ( String.sub
					    (lexeme lexbuf)
					    1
					    ( String.length (lexeme lexbuf) - 2 ) )
			}
| '"' [' ''!''#'-'~']* '"'
			{ STRING ( String.sub
				     (lexeme lexbuf)
				     1
				     ( String.length (lexeme lexbuf) - 2 ) )
			}
| eof			{ EOF }
| _			{ Printf.printf "Unexpected character in line %d : %s\n"
			    !linecount
			    (Lexing.lexeme lexbuf);
			  flush stdout;
			  failwith "Lexical error." }
