(* Franois Thomasset -- INRIA Rocquencourt -- Octobre 2001 *)
(*
Copyright  2001-2002 Franois Thomasset, all rights reserved.
Copying is covered by the GNU General Public License (GPL).
 
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
 
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.
 
$Revision: 1.1.1.1 $
$Date: 2002/03/08 08:31:57 $

*)
open Maple;;
open Kel;;
open Util;;
open Stringtable;;

(* simplifier l'expression (e+i) avec i constante entire *)
let add_int_to_expr i = function
  | IntExpr (n) -> IntExpr(n+i)
  | BinExpr ({ left = l; bin_op = the_op ; right = r }) as binexp ->
      let addi () =
	if ( i = 0 ) then binexp
	else if ( i > 0 )
	then BinExpr ({ left = binexp; bin_op = PlusOp;  right = IntExpr(i) })
	else BinExpr ({ left = binexp; bin_op = MinusOp; right = IntExpr(i) })
      in
      begin
	match the_op with
	| PlusOp ->
	    begin
	      match l with
	      |	IntExpr (li) ->
		  if ( (li+i) = 0 )
		  then r
		  else BinExpr ({ left = IntExpr (li+i); bin_op = the_op ; right = r })
	      |	_ ->
		  begin
		    match r with
		    | IntExpr (ri) ->
			if ( (ri+i) = 0 )
			then l
			else BinExpr ({ left = l; bin_op = the_op ; right = IntExpr (ri+i) })
		    | _ -> addi()
		  end
	    end
	| MinusOp ->
	    begin
	      match l with
	      |	IntExpr (li) ->
		  if ( (li-i) = 0 )
		  then UnaryExpr ( UnaryMinus , r )
		  else BinExpr ({ left = IntExpr (li-i); bin_op = the_op ; right = r })
	      |	_ ->
		  begin
		    match r with
		    | IntExpr (ri) ->
			if ( (ri+i) = 0 ) then l
			else BinExpr ({ left = l; bin_op = the_op ; right = IntExpr (ri+i) })
		    | _ -> addi()
		  end
	    end
	| _ -> addi()
      end
  | e ->
      if ( i = 0 ) then e
	else if ( i > 0 )
	then BinExpr ({ left = e; bin_op = PlusOp;  right = IntExpr(i) })
	else BinExpr ({ left = e; bin_op = MinusOp; right = IntExpr(-i) });;

(* The big translation function *)

let prt_prog out_ch evaluation_level verbose le_programme =
   (* current_col sert  marquer la position courante dans la ligne de sortie.
     Pour que cela reste vrai, il est impratif de respecter les rgles suivantes :
     1. Pour mettre une chane sans fin de ligne, utiliser say.
     2. Pour mettre une chane avec fin de ligne, utiliser sayln
       (qui remet  zro la position courante)
     Pour mettre un entier on peut utiliser say_int.
   *)
  let current_col = ref 0 and current_indentation = ref 0 in
  let max_col = 90 in
  let say_string s = output_string out_ch s in
  let say s = begin say_string s; current_col := !current_col + (String.length s) end in
  let say_int i = say (string_of_int i) in
  let sayln s = begin say_string s; say_string "\n"; current_col := 0 end in
  let indent s =
    let rec indent_rec indentation_factor =
    if ( indentation_factor > 0 )
    then begin say_string " "; indent_rec (indentation_factor-1) end
    else ()
    in begin indent_rec s; current_col := !current_col + s; current_indentation := s end
  in
  (* On imprime les commentaires sans indenter : ncessaire pour la documentation *)
  let rec prt_comments debut = function
    | [] -> ()
    | c :: suite -> begin if (debut) then sayln ""; say "//"; sayln c; prt_comments false suite end
  in

  let say_op = function
    | PlusOp -> say " + " | MinusOp -> say " - "
    | MultOp -> say " * " | DivOp -> say " / "
    | ExpoOp -> say " ^ "
    | OrOp -> say " or "  | AndOp -> say " and "
    | LTOp -> say " < "   | GTOp -> say " > "
    | LEOp -> say " <= "  | GEOp -> say " >= "
    | NEOp -> say " <> "  | EQOp -> say " = "
    | ModOp -> say " mod "
    | NeutralMulOp -> say " &* "
    | NeutralOp (o) -> begin say " &"; say o; say " " end
    | Union -> say " union " | Minus -> say " minus " | Intersect -> say " intersect "
  in
  let prt_op oper =
    if ( !current_col > max_col )
    then begin sayln ""; indent (!current_indentation); say "   " end;
    say_op oper
  in
  let prt_unary_op = function
    | UnaryMinus -> say " - " | UnaryPlus -> say " + " | NotOp -> say " not "
  in
  (* Pour imprimer les noms *)
  let match_none_name = function
    | Some e -> false
    | _ -> true
  and match_none_expr = function
    | Some e -> false
    | _ -> true
  in
  let change_keywords = function
    | "assign" -> "__assign"         | "is" -> "__is"
    | "expr" -> "__expr"             | "hold" -> "__hold"
    | "left" -> "__left"             | "right" -> "__right"
    | "true" -> "TRUE"               | "false" -> "FALSE"
    | "TRUE" -> "__TRUE"             | "FALSE" -> "__FALSE"
    | "NULL" -> "null()"             | "null" -> "__null"
    | "Order" -> "ORDER"             | "screenwidth" -> "TEXTWIDTH"
    | "Digits" -> "DIGITS"
    | "Pi" -> "PI"                   | "PI" -> "__PI"
    | "gamma" -> "EULER"             | "EULER" -> "__EULER"
    | "nargs" -> "args(0)"
    | n -> n
  in
  let prt_name_string = function
    | Ident (n) -> say (change_keywords n)
    | QuotedString ("*") -> say "hold(_mult)"
    | QuotedString ("+") -> say "hold(_plus)"
    | QuotedString ("^") -> say "hold(_power)"
    | QuotedString ("=") -> say "hold(_equal)"
    | QuotedString (">=") -> say "hold(_geequal)"
    | QuotedString ("<=") -> say "hold(_leequal)"
    | QuotedString (">") -> say "hold(_less)"
    | QuotedString ("<") -> say "hold(_less)"
    | QuotedString (s) -> begin say "\""; say s; say "\"" end
  in
  let rec prt_name = function
    | NameString (n) -> prt_name_string n
    | NameIndexed (n,el) ->
	begin
	  match n with
	  | NameString(Ident("args")) | NameString(QuotedString("args")) ->
	      begin
		prt_name n;
		say "("; prt_exprs_no_paren el; say ")"
	      end
	  | _ ->
	      begin
		prt_name n;
		say "["; prt_exprs_no_paren el; say "]"
	      end
	end
    | NameConcat (n,nc)
      ->
	begin
	  prt_name n; say " . ";
	  match nc with
	  | N_int (i) -> say_int i
	  | N_str (s) -> say s
	  | N_expr (e) -> begin say "( "; prt_expr e; say ")" end
	end
  and prt_paramseq = function
    | [] -> ()
    | [ { param_name = n ; param_type = ty } ] ->
	begin
	  prt_name n;
	  match ty with
	  | Some tyname -> begin say ":"; prt_typename tyname end
	  | _ -> ()
	end
    | { param_name = n ; param_type = ty } :: suite ->
	begin
	  prt_name n;
	  begin
	    match ty with
	    | Some tyname -> begin say ":"; prt_typename tyname end
	    | _ -> ()
	  end;
	  say ",";
	  if ( !current_col > max_col )
	  then
	    begin
	      sayln ""; indent !current_indentation; say "      ";
	    end;
	  prt_paramseq suite
	end

  and prt_nameseq = function
    | [] -> ()
    | [n] ->  prt_name n
    | n :: suite ->
	begin
	  prt_name n; say ",";
	  if ( !current_col > max_col )
	  then
	    begin
	      sayln "";
	      indent !current_indentation;
	      say "      ";
	    end;
	  prt_nameseq suite
	end

  (* Impression d'une liste d'expressions *)
  and prt_suite_exprs = function
    | [] -> ()
    | a :: suite -> begin say ", "; prt_expr a; prt_suite_exprs suite end
  and prt_exprs = function
    | [] -> say " ()"
    | [a] -> begin say " ( "; prt_expr a; say " ) " end
    | a :: suite -> begin say " ( "; prt_expr a; prt_suite_exprs suite; say " ) " end
  (* sans parenthses en dbut et fin de liste *)
  and prt_exprs_no_paren = function
    | [] -> ()
    | [a] -> prt_expr a
    | a :: suite -> begin prt_expr a; prt_suite_exprs suite end

  and get_name_string = function
    | NameString (n) ->
	begin match n with Ident (x) -> x | _ -> "" end
    | _ -> ""

  and string_of_name = function
    | NameString (n) ->
	begin
	  match n with
	  | Ident (x) -> x
	  | QuotedString (x) -> ("`" ^ x ^ "`")
	end
    | NameIndexed (n,_) -> (string_of_name n) ^ "[ ...]"
    | NameConcat (n,_) -> (string_of_name n) ^ " . ( ... )"
  
  (* utilitaire pour traduire les appels  map, select *)
  and prt_map_like fname fargs =
    match fargs with
    | [] -> say fname; say "()"
    | [x] -> begin say fname; say "("; prt_expr x; say ")" end
    | obj :: f :: reste ->
	begin
	  say fname; say "(";
	  prt_expr f; say ","; prt_expr obj;
	  begin
	    match reste with
	    | [] -> ()
	    | _ -> begin say ","; prt_exprs_no_paren reste end
	  end;
	  say ")"
	end
  (* Imprimer une expression *)
  and prt_expr = function
    | ArrowExpr (n,e) ->
	begin
	  begin
	    match n with
	    | [] ->
		begin
		  Printf.printf "WARNING: arrow expr with empty name list";
		  say "()"
		end
	    | [n] -> prt_name n
	    | nl -> begin say "("; prt_nameseq nl; say ")" end ;
	  end;
	  say " -> "; prt_expr e
	end
    | BinExpr (e) as binexp ->
	begin
	  match binexp with
	  | BinExpr ({ left = NameExpr(NameIndexed(n1,[IntervalExpr(l1,r1)])) ;
		       bin_op = the_op ;
		       right = NameExpr(NameIndexed(n2,[IntervalExpr(l2,r2)]))
		     })
	    ->
	      begin
		let make_new_index () = "__" ^ (get_unique_str()) ^ "__index__"
		and make_new_var() = "__" ^ (get_unique_str()) ^ "__e__"
		in
		let indx1 = make_new_index() and indx2 = make_new_index()
		in
		let e1 = make_new_var() and e2 = make_new_var()
		in
		say "(";
		say e1; say ":=";
		prt_expr
		  (SeqExpr
		     (
		      Some (NameExpr(NameIndexed(n1,[NameExpr(NameString(Ident(indx1)))]))),
		      BinExpr ({ bin_op = EQOp;
				 left = NameExpr(NameString(Ident(indx1)));
				 right = IntervalExpr(l1,r1)
			       }
			      )
		     )
		  );
		say ":";
		say e2; say ":=";
		prt_expr
		  (SeqExpr
		     (
		      Some (NameExpr(NameIndexed(n2,[NameExpr(NameString(Ident(indx2)))]))),
		      BinExpr ({ bin_op = EQOp;
				 left = NameExpr(NameString(Ident(indx2)));
				 right = IntervalExpr(l2,r2)
			       }
			      )
		     )
		  );
		say ":";
		say "if (testtype("; say e1; say ",DOM_NULL) or ";
		say "testtype("; say e2; say ",DOM_NULL))";
		say "then null() ";
		say "else ("; say e1; prt_op the_op; say e2; say ")";
		say " end_if";
		say ")"
	      end
	  | _ ->
	      begin
		prt_expr e.left; prt_op e.bin_op; prt_expr e.right
	      end
	end
    | Factorial (e) -> begin prt_expr e; say "!" end
    | UnaryExpr (op,e) -> begin prt_unary_op op; say " "; prt_expr e end
    | SeqExpr (i,e) ->
	(* Try to match f(s[i]) $ (i = 1 .. nops(s))
	   and generate: map(s,f)
	   Check for the case of non evaluated sub expressions:
	   'f(s[i])' $ ('i' = 1 .. nops(s))
	   *)
	let rec chk_indx i maybe_an_index_name maybe_a_set =
	  match i with
	  | Some CallExpr ( { callee = NameString ( Ident ( maybe_a_fun ) ) ;
			      args = [ NameExpr
					 ( NameIndexed ( NameString (Ident (a_set)),
							 [ NameExpr (NameString(Ident(an_index)))]
							)
					  )
				     ]
			    }
			   )
	    when ((maybe_an_index_name = an_index) &&
	          (maybe_a_set = a_set))
	    -> Some (maybe_an_index_name, maybe_a_set, maybe_a_fun)
	  | Some (UnevaluatedExpr 
		    ( CallExpr ( { callee = NameString ( Ident ( maybe_a_fun ) ) ;
				   args = [ NameExpr
					      ( NameIndexed ( NameString (Ident (a_set)),
							      [ NameExpr (NameString(Ident(an_index)))]
							     )
					       )
					  ]
				 }
				)
		     )
		 )
	    when ((maybe_an_index_name = an_index) &&
	          (maybe_a_set = a_set))
	    -> Some (maybe_an_index_name, maybe_a_set, maybe_a_fun)
	  | Some (UnevaluatedExpr ( ii )) -> chk_indx (Some ii) maybe_an_index_name maybe_a_set
	  |  _ -> None
	in
	let rec chk_map i e =
	  match e with
	  | BinExpr ( { bin_op = EQOp;
			left = NameExpr(NameString(Ident(maybe_an_index_name)));
			right = IntervalExpr(IntExpr(1),
					     CallExpr ( { callee =
							  NameString
								( Ident ( "nops" ) ) ;
							  args = [ NameExpr
								     ( NameString
									 ( Ident ( maybe_a_set ) ) )
								 ]
							}
						       )
					    )
		      }
		     )
	    -> chk_indx i maybe_an_index_name maybe_a_set
	  | BinExpr ( { bin_op = EQOp;
			left = UnevaluatedExpr(NameExpr(NameString(Ident(maybe_an_index_name))));
			right = IntervalExpr(IntExpr(1),
					     CallExpr ( { callee =
							  NameString
								( Ident ( "nops" ) ) ;
							  args = [ NameExpr
								     ( NameString
									 ( Ident ( maybe_a_set ) ) )
								 ]
							}
						       )
					    )
		      }
		     )
	    -> chk_indx i maybe_an_index_name maybe_a_set
	  | ParenExpr ( [ee] ) -> chk_map i ee
	  | UnevaluatedExpr ( ee ) -> chk_map i ee
	  | _ -> None
	in
	let indx_set_fun = chk_map i e
	in
	begin
	  match indx_set_fun with
	  | Some (indx,a_set,a_fun)
	    ->
	      begin
		say "map(";
		say a_set;
		say ",";
		prt_expr (ArrowExpr([NameString(Ident(indx))],
				    CallExpr ( { callee=NameString(Ident(a_fun));
						 args = [NameExpr(NameString(Ident(indx)))]
					       } )));
		say ")"
	      end
	  | _ ->
	      begin
		begin
		  match i with
		  | Some ii -> begin prt_expr ii; say " " end
		  | _ -> ()
		end ;
		say "$ ";
		prt_expr e
	      end
	end
    | IntervalExpr (a,b) -> begin prt_expr a; say " .. "; prt_expr b end
    | NameExpr (n) -> prt_name n
    | StringExpr (s) -> begin say "\""; say s; say "\""; end
    | FloatExpr (i, f) -> begin say_int i; say "."; say_int f end
    | IntExpr (n) -> say_int n
    | CallExpr (ce)
      ->
	let prt_call_expr_general_case () =
	  prt_name ce.callee; prt_exprs ce.args
	in
	begin
	  match get_name_string ce.callee with
	  | "signum" -> begin say "sign"; prt_exprs ce.args end
	  | "substring" ->
	      let prt_moins_un e = prt_expr (add_int_to_expr (-1) e)
	      in
	      begin
		match ce.args with
		| [ str ; IntervalExpr (low,high) ] ->
		    begin
		      say "substring(";
		      prt_expr str; say ",";
		      prt_moins_un low; say ","; prt_moins_un high;
		      say ")"
		    end
		| [ str ; i ] ->
		    begin
		      say "substring(";
		      prt_expr str; say ",";
		      prt_moins_un i; say ")"
		    end
		| _ -> prt_call_expr_general_case()
	      end
	  | "simplify" ->
	      begin
		match ce.args with
		| [ f ] -> prt_call_expr_general_case()
		| [ f ; rules ] ->
		    begin say "subs"; prt_exprs ce.args end
		| _ -> prt_call_expr_general_case()
	      end
	  | "subs" ->
	      begin
		say "subs(";
		begin
                 match ce.args with
		  | [] -> ()
		  | [x] -> prt_expr x
		  | _ ->
		      begin
			let reverse_list_args = List.rev ce.args
			in
			prt_expr (List.hd reverse_list_args);
			say ",";
			prt_exprs (List.rev (List.tl reverse_list_args))
		      end
		end ;
		say ")"
	      end
	  | "subsop" ->
	      begin
		say "subsop(";
		begin
                 match ce.args with
		  | [] -> ()
		  | [x] -> prt_expr x
		  | _ ->
		      begin
			let reverse_list_args = List.rev ce.args
			in
			prt_expr (List.hd reverse_list_args);
			say ",";
			prt_exprs (List.rev (List.tl reverse_list_args))
		      end
		end ;
		say ")"
	      end
	  | "op" ->
	      begin
		say "op(";
		begin
		  match ce.args with
		  | [] -> ()
		  | [x] -> prt_expr x
		  | [x;y] -> begin prt_expr y; say " , "; prt_expr x end
		  | _ ->
		      begin
			Printf.printf "WARNING: op with more than 2 args\n";
			prt_exprs ce.args
		      end
		end;
		say ")"
	      end
	  | "member" ->
	      begin
		match ce.args with
		| [x;f] ->
		    begin
		      say "contains (";
		      say "{op("; prt_expr f; say ")}";
		      say ","; prt_expr x; say ")"
		    end
		| _ -> prt_call_expr_general_case()
	      end
	  | "sort" ->
	      begin
		match ce.args with
		| [x] ->
		    begin
		      say "sort("; say "[op("; prt_expr x; say ")], maple_order) "
		    end
		| _ -> prt_call_expr_general_case()
	      end
	  | "RETURN" -> begin say "return "; prt_exprs ce.args end
	  | "ERROR" ->
	      begin
		say "error ";
		let arg_to_str = function
		  | NameExpr(NameString (QuotedString(n))) ->
		      begin say "\""; say n; say "\""; end
		  | StringExpr(s) ->
		      begin say "\""; say s; say "\""; end
		  | a -> begin say "expr2text("; prt_expr a; say ")" end
		in
		begin
		  match ce.args with
		  | [] -> ()
		  | [ arg ] -> begin say "("; arg_to_str arg; say ")" end
		  | _ ->
		      let rec error_args_to_string = function
			| [] -> ()
			| [ arg ] -> arg_to_str arg
			| arg :: suite ->
			    begin
			      arg_to_str arg; say " . "; error_args_to_string suite
			    end
		      in
		      begin say "("; error_args_to_string ce.args; say ")" end
		end
	      end
	  | "printf" ->
	      begin
		match ce.args with
		  | [] -> ()
		  | format_string :: reste_args ->
		      prt_printf_expr format_string reste_args
	      end
	  | "fprintf" ->
	      begin
		match ce.args with
		  | [] -> ()
		  | [x] ->
		      begin
			Printf.printf "WARNING: fprintf with only 1 arg\n";
			prt_expr x
		      end
		  | file_descriptor :: format_string :: reste_args ->
		      prt_fprintf_expr file_descriptor format_string reste_args
	      end
	  | "sprintf" ->
	      begin
		match ce.args with
		  | [] -> say "\"\""
		  | [x] -> prt_expr x
		  | StringExpr(fmt) :: reste_args -> prt_sprintf fmt reste_args
		  | fmt  :: reste_args ->
		      begin
			Printf.printf "WARNING: 1st arg of sprintf is not a string\n";
			say "sprintf"; prt_exprs ce.args
		      end
	      end
	  | "fopen" ->
	      begin
		(* Pour crire en mode texte, il faut donner Text comme 1er argument  fopen *)
		match ce.args with
		  | [] ->
		      begin
			Printf.printf "WARNING: fopen with an empty list of args\n";
			say "fopen () "
		      end;
		  | [filename;mode] ->
		      begin
			match mode with
			| NameExpr (NameString(Ident("WRITE"))) ->
			    begin
			      say "fopen(Text,"; prt_expr filename; say ", Write)"
			    end
			| _ ->
			    begin
			      say "fopen("; prt_expr filename; 
			      begin
				match mode with
				| NameExpr (NameString(Ident("APPEND"))) -> say ",Append"
				| NameExpr (NameString(Ident("READ"))) -> ()
				| NameExpr (NameString(Ident(unknown_mode))) ->
				    begin
				      Printf.printf "WARNING: fopen with unknown mode: %s\n" unknown_mode;
				      say ","; say unknown_mode
				    end
				| _ ->
				    begin
				      Printf.printf "WARNING: fopen with unexpected mode: not a NameExpr\n";
				      say ","; prt_expr mode
				    end
			      end ;
			      say ")"
			    end
		      end
		  | _ ->
		      begin
			Printf.printf "WARNING: fopen with more than 2 args\n";
			say "fopen "; prt_exprs ce.args
		      end
	      end
	  | "readline" ->
		begin
		  match ce.args with
		  | [ fd ] -> begin say "ftextinput("; prt_expr fd; say ")" end
		  | _ -> begin prt_name ce.callee; prt_exprs ce.args end
		end
	  | "type" ->
	      begin
		begin
		  match ce.args with
		  | [obj ; t] ->
		      begin
			if ( (is_this_type "constant" t) or (is_this_type "numeric" t) )
			then
			  begin
			    let prt_type_cst ty =
			       say " (testtype ("; prt_expr obj; say ","; say ty; say "))"
			    in
			    prt_type_cst "DOM_INT"; say " or ";
			    prt_type_cst "DOM_RAT"; say " or ";
			    prt_type_cst "DOM_FLOAT"
			  end
			else if (is_this_type "name" t)
			then
			  begin say " is_name ("; prt_expr obj; say ")" end
			else if (is_this_type "linear" t)
			then
			  begin say "is_linear ("; prt_expr obj; say ")" end
			else if (is_this_type "relation" t)
			then
			  begin say "is_relation ("; prt_expr obj; say ")" end
			else if (is_this_type "equation" t)
			then
			  begin say "is_equation ("; prt_expr obj; say ")" end
			else if (is_this_type "algebraic" t)
			then
			  begin say "is_algebraic ("; prt_expr obj; say ")" end
			else
			  begin
			    say " testtype (";
			    prt_expr obj; say ","; prt_type t ;
			    say ")"
			  end
		      end
		  | _ -> begin Printf.printf "WARNING: wrong number of args for type" end
		end
	      end
	  | "whattype" -> begin say " type "; prt_exprs ce.args end
	  | "table" ->
	      begin
		say "table ";
		begin
		  match ce.args with
		  | [SetExpr (es)] -> prt_exprs es
		  | [] -> say "()"
		  | _ ->
		      begin
			Printf.printf "WARNING: argument of table not a SetExpr?\n";
			say "("; prt_exprs ce.args; say ")"
		      end
		end
	      end
	  | "eval" ->
	      begin
		match ce.args with
		| [] ->
		    begin say "eval"; prt_exprs ce.args end
		| [e ; IntExpr(l)] ->
		    begin say "level("; prt_expr e; say ","; say_int l; say ")" end
		| [e] ->
		    begin say "eval"; prt_exprs ce.args end
		| _ ->
		    begin say "subs"; prt_exprs ce.args end
	      end
	  | "evalb" ->
	      begin say "bool"; prt_exprs ce.args; end
	  | "evalf" ->
	      begin say "float"; prt_exprs ce.args; end
	  | "evaln" ->
	      begin say "hold"; prt_exprs ce.args; end
	  | "assigned" ->
	      begin
		match ce.args with
		| [x] -> begin say "contains(anames(All), hold("; prt_expr x; say "))" end
		| _ ->begin say "assigned"; prt_exprs ce.args; end
	      end
	  | "coeffs" ->
	      begin
		match ce.args with
		| [poly ; v] ->
		    begin say "coeff("; prt_expr poly; say ",["; prt_expr v; say "]" end
		| [poly] ->
		    begin say "coeff("; prt_expr poly; say ")" end
		| _ -> 
		    begin
		      Printf.printf "WARNING: in Maple coeffs expects 1 or 2 arguments\n";
		      Printf.printf "But here coeffs has %d arguments\n" (List.length ce.args);
		      say "coeffs"; prt_exprs ce.args
		    end
	      end
	  | "coeff" ->
	      begin
		match ce.args with
		| [poly ; v] ->
		    begin
		      say "maple_coeff("; prt_expr poly; say ","; prt_expr v; say ")"
		    end
		| [poly ; v ; i ] ->
		    begin
		      say "coeff(";
		      prt_expr poly; say ","; prt_expr v; say ","; prt_expr i;
		      say ")"
		    end
		| _ -> 
		    begin
		      Printf.printf "WARNING: in Maple coeff expects 2 or 3 arguments\n";
		      Printf.printf "But here coeffs has %d arguments\n" (List.length ce.args);
		      say "coeff"; prt_exprs ce.args
		    end
	      end
	  | "cat" ->
	      begin
		match ce.args with
		  | [] -> ()
		  | [x] -> prt_expr x
		  | x :: y ->
		      let rec prt_cat_args = function
			| [] -> ()
			| x :: y -> begin say " . "; prt_expr x; prt_cat_args y end
		      in
		      begin prt_expr x; prt_cat_args y end
	      end
	  | "log" -> begin say "ln"; prt_exprs ce.args end
	  | "GAMMA" -> begin say "gamma"; prt_exprs ce.args end
	  (* pour select, zip et map,
	     il faut faire passer le nom de la fonction en 2eme position *)
	  | "select" -> prt_map_like "select" ce.args
	  | "map" -> prt_map_like "map" ce.args
	  | "zip" ->
	      begin
		match ce.args with
		| [ foo ; a ; b ] ->
		    begin
		      say "zip(";
		      prt_expr a; say ",";
		      prt_expr b; say ",";
		      prt_expr foo; say ")"
		    end
		| _ -> begin prt_name ce.callee; prt_exprs ce.args end
	      end
	  | "convert" ->
	      begin
		match ce.args with
		| [ ListExpr (le) as l; NameExpr(NameString(operation_name)) ] ->
		    begin
		      match operation_name with
		      |	QuotedString("+") ->
			  begin say "_plus(op("; prt_expr l; say "))" end
		      |	QuotedString("*") ->
			  begin say "_mult(op("; prt_expr l; say "))" end
		      |	Ident("set") | QuotedString("set") ->
			  begin say "{"; prt_exprs_no_paren le; say "}" end
		      |	_ -> begin prt_name ce.callee; prt_exprs ce.args end
		    end
		| [x ; NameExpr(NameString(s))] ->
		    begin
		      match s with
		      |	Ident("set") | QuotedString("set") ->
			  begin say "{op("; prt_expr x; say ")}" end
		      |	Ident("list") | QuotedString("list") ->
			  begin say "[op("; prt_expr x; say ")]" end
		      |	_ -> prt_call_expr_general_case()
		    end
		| _ -> prt_call_expr_general_case()
	      end
	  | "array" ->
	      begin
		match ce.args with
		| [ NameExpr(NameString(Ident("symmetric"))) ; _ ]
		| [ NameExpr(NameString(QuotedString("symmetric"))) ; _ ] ->
		    begin
		      Printf.printf "WARNING: I cannot translate the symmetric option in array\n";
		      prt_name ce.callee; prt_exprs ce.args
		    end
		| _ ->
		    begin
		      match (List.rev ce.args) with
		      |	ListExpr (ldef) :: rdims ->
			  begin
			    say "array(";
			    prt_exprs_no_paren (List.rev rdims);
			    say ","; prt_exprs_no_paren ldef; say ")"
			  end
		      |	_ -> prt_call_expr_general_case()
		    end
	      end
	  | "Int" ->
	      begin
		match ce.args with
		| [ f ; x ] ->
		    begin say "hold(int("; prt_expr f; say ","; prt_expr x; say "))" end
		| _ -> prt_call_expr_general_case()
	      end
	  | "limit" | "Limit" ->
	      begin
		let fn_is_Limit = ( (get_name_string ce.callee) = "Limit")
		in
		match ce.args with
		| [ e ; v ] ->
		    begin
		      if fn_is_Limit then say "hold(";
		      say "limit"; prt_exprs ce.args;
		      if fn_is_Limit then say ")"
		    end
		| [ e ; v ; opt ] ->
		    begin
		      let prt_opt = function
		      |	NameExpr(NameString(Ident("left")))
		      |	NameExpr(NameString(QuotedString("left"))) ->
			  say "Left"
		      |	NameExpr(NameString(Ident("right")))
		      |	NameExpr(NameString(QuotedString("right"))) ->
			  say "Right"
		      |	_ -> prt_expr opt
		      in
		      begin
			if fn_is_Limit then say "hold(";
			say "limit("; prt_expr e; say","; prt_expr v;
			say ","; prt_opt opt; say ")";
			if fn_is_Limit then say ")"
		      end
		    end
		| _ -> prt_call_expr_general_case()
	      end
	  | "newvar" ->
	      (* c'est notre propre fonction pour fabriquer des nouveaux noms *)
	      begin
		match ce.args with
		| [] -> say "genident()"
		| [NameExpr(NameString(Ident(base)))] ->
		    begin say "genident(\""; say base; say "\")" end
		| [NameExpr(NameString(QuotedString(base)))] ->
		    begin say "genident(\""; say base; say "\")" end
		| _ -> prt_call_expr_general_case()
	      end
	  | "newvars" ->
	      begin
		match ce.args with
		| [NameExpr(NameString(Ident(base))) ; lst] ->
		    begin say "newvars(\""; say base; say "\","; prt_expr lst; say ")" end
		| _ -> prt_call_expr_general_case()
	      end
          (* Si on n'a pas reconnu de nom de fonction, cas gnral *)
	  | _ ->
	      begin
		let general_case () =
		  if (verbose)
		  then addname (string_of_name ce.callee);
		  prt_call_expr_general_case()
		in
		match ce.callee with
		| NameIndexed (n,[f]) ->
		    begin
		      if (verbose)
		      then addname (string_of_name ce.callee);
		      prt_name n; say "::"; prt_expr f; prt_exprs ce.args
		    end
		| _ -> general_case ()
	      end
	end

    | FunctionalCallExpr (ce)
      ->
	let rec prt_all_args = function
	  | [] -> ()
	  | a :: suite ->
	      begin prt_exprs a; prt_all_args suite end
	in
	begin prt_name ce.callee_f; prt_all_args ce.fargs end

    | UnevaluatedExpr (e) -> begin say "hold( "; prt_expr e; say " )" end
    | ListExpr (es) -> begin say " [ "; prt_exprs_no_paren es; say " ] " end
    | SetExpr (es) -> begin say " { "; prt_exprs_no_paren es; say " } " end
    | ProcDef (p)
      ->
	  begin
	    say "proc ( "; prt_paramseq p.params; say " )";
	    begin
	      match p.result_type with
	      |	Some n -> begin say " : "; prt_typename n end
	      |	_ -> ()
	    end;
	    begin
	      match p.locals with
	      | [] -> ()
	      | _ -> begin sayln ""; indent 3; say "local "; prt_nameseq p.locals; say ";" end
	    end;
	    begin
	      let rec rm_operator = function
		| [] -> []
		| NameString(Ident("operator")) :: suite -> rm_operator suite
		| NameString(QuotedString("operator")) :: suite -> rm_operator suite
		| x :: suite -> x :: (rm_operator suite)
	      in
	      let opts = (rm_operator p.options)
	      in
	      match opts with
	      |	[] -> ()
	      | _ ->
		  begin
		    sayln "";
		    indent 3; say "option "; prt_nameseq opts;
		    say ";"
		  end
	    end;
	    sayln ""; indent 3; sayln "begin";
	    prt_stmts 6 true p.body_of_proc; sayln "";
	    indent 3; say "end_proc"
	  end
    | ParenExpr (e) -> begin say "("; prt_exprs e; say ")" end
    | ComposeExpr (f,g) ->  begin prt_expr f; say "@"; prt_expr g end
    | RepeatComposeExpr (f,i) ->  begin prt_expr f; say "@@"; prt_expr i end

  (* pour la fonction type *)
  and is_this_type a_type = function
    | NameExpr ( NameString ( Ident (my_type) ) ) -> (a_type = my_type)
    | NameExpr ( NameString ( QuotedString (my_type) ) ) -> (a_type = my_type)
    | UnevaluatedExpr ( t ) -> is_this_type a_type t
    | _ -> false
  and prt_type = function
    | NameExpr (NameString(typename)) -> prt_typename typename
    | UnevaluatedExpr (e) -> prt_type e
    | e -> prt_expr e
  and prt_typename = function
    | Ident ("float") -> say "DOM_FLOAT"
    | QuotedString ("float") -> say "DOM_FLOAT"
    | Ident ("integer") -> say "DOM_INT"
    | QuotedString ("integer") -> say "DOM_INT"
    | Ident ("list") -> say "DOM_LIST"
    | QuotedString ("list") -> say "DOM_LIST"
    | Ident ("name") -> say "DOM_IDENT"
    | QuotedString ("name") -> say "DOM_IDENT"
    | Ident ("fraction") -> say "DOM_RAT"
    | QuotedString ("fraction") -> say "DOM_RAT"
    | Ident ("rational") -> say "DOM_RAT"
    | QuotedString ("rational") -> say "DOM_RAT"
    | Ident ("complex") -> say "DOM_COMPLEX"
    | QuotedString ("complex") -> say "DOM_COMPLEX"
    | Ident ("symbol") -> say "DOM_IDENT"
    | QuotedString ("symbol") -> say "DOM_IDENT"
    | Ident ("set") -> say "DOM_SET"
    | QuotedString ("set") -> say "DOM_SET"
    | Ident ("indexed") -> say "\"_index\""
    | QuotedString ("indexed") -> say "\"_index\""
    | Ident ("function") -> say "\"function\""
    | QuotedString ("function") -> say "\"function\""
    | QuotedString ("*") -> say "\"_mult\""
    | QuotedString ("/") -> say "\"_mult\""
    | QuotedString ("+") -> say "\"_plus\""
    | QuotedString ("-") -> say "\"_plus\""
    | QuotedString ("^") -> say "\"_power\""
    | QuotedString ("=") -> say "\"_equal\""
    | QuotedString (">=") -> say "\"_leequal\""
    | QuotedString ("<=") -> say "\"_leequal\""
    | QuotedString (">") -> say "\"_less\""
    | QuotedString ("<") -> say "\"_less\""
    | n -> prt_name_string n

  and prt_printf_expr format_string reste_args =
    let fmt =
      match format_string with 
      |	StringExpr(s) -> s
      | _ -> begin Printf.printf "WARNING:printf with no format string\n"; "" end
    in
    let keep_indent = !current_indentation and keep_curpos = !current_col
    in
    let prt_print_call nl_or_nonl =
      say "print("; say nl_or_nonl 
    in
    prt_print_expr prt_print_call fmt keep_curpos reste_args

  (* traduction de l'appel  fprintf : ce n'est pas le code dont je suis le plus fier *)
  and prt_fprintf_expr file_descriptor format_string reste_args =
    let fmt =
      match format_string with 
      |	StringExpr(s) -> s
      | _ -> begin Printf.printf "WARNING: fprintf with no format string\n"; "" end
    in
    let keep_indent = !current_indentation and keep_curpos = !current_col
    in
    let prt_print_call nl_or_nonl =
      say "fprint("; say nl_or_nonl; say ",";
      prt_expr file_descriptor
    in
    prt_print_expr prt_print_call fmt keep_curpos reste_args

  and prt_print_expr prt_print_call fmt keep_curpos reste_args =
    let len = (String.length fmt)
    in
    let rec prt_format_string current_pos =
      (* traduction d'un appel  fprintf avec format simple, sans arguments supplmentaires *)
      (* on teste si l'on doit rcurser avant d'entrer dans la rcursion :
	 permet de ne pas mettre de point-virgule aprs le dernier fprint
	 (le point-virgule sera mis  la fin de prt_stat) *)
      if (len <= current_pos) then ()
      else
	try let pos_nl = find_slash_n current_pos fmt
	in
	begin
	  prt_print_call "Unquoted"; say ",\"";
	  say (String.sub fmt current_pos (pos_nl - current_pos));
	  say "\")";
	  if (len > (pos_nl + 2))
	  then begin sayln " ;"; indent keep_curpos; prt_format_string (pos_nl + 2) end
	end
	with Not_found ->
	  begin
	    prt_print_call "NoNL"; say ",\"";
	    say (String.sub fmt current_pos (String.length fmt - current_pos));
	    say "\") "
	  end
    in
    let must_recurse pos = function
      |	[] -> (len > pos)
      |	_ -> true
    in
    (* on teste si l'on doit rcurser avant d'entrer dans la rcursion :
       permet de ne pas mettre de point-virgule aprs le dernier fprint
       (le point-virgule sera mis  la fin de prt_stat) *)
    let rec prt_args_of_print current_pos = function
      | [] -> prt_format_string current_pos
      | a :: reste_args as liste_args ->
	  begin
	    try let pos_percent = (String.index_from fmt current_pos '%')
	    in
	    begin
	      try let pos_nl = find_slash_n current_pos fmt
	      in
	      if (pos_nl < pos_percent)
	      then
		begin
		  prt_print_call "Unquoted"; say ",\"";
		  say (String.sub fmt current_pos (pos_nl - current_pos));
		  say "\")";
		  if (len > (pos_nl + 2))
		  then
		    begin
		      sayln " ;";
		      indent keep_curpos;
		      prt_args_of_print (pos_nl + 2) liste_args
		    end
		end
	      else
		begin
		  prt_print_call "NoNL"; say ",";
		  if (current_pos < pos_percent)
		  then
		    begin
		      if (len = (pos_percent+1))
		      then
			begin
			  say "\"";
			  say (String.sub fmt current_pos (pos_percent - current_pos + 1));
			  say "\","
			end
		      else if (pos_percent > current_pos) then
			begin
			  say "\"";
			  say (String.sub fmt current_pos (pos_percent - current_pos));
			  say "\","
			end;
		    end ;
		  prt_expr a;
		  sayln ") ;"; indent keep_curpos;
		  if (len = (pos_percent+1))
		  then prt_args_of_print (pos_percent+1) reste_args
		  else prt_args_of_print (pos_percent+2) reste_args
		end
	      with Not_found | Invalid_argument _ ->
		begin
		  prt_print_call "NoNL"; say ",";
		  if (current_pos < pos_percent)
		  then
		    begin
		      say "\"";
		      say (String.sub fmt current_pos (pos_percent - current_pos));
		      say "\",";
		    end;
		  prt_expr a;
		  say ") ;";
		  if (must_recurse (pos_percent+2) reste_args)
		  then
		    begin
		      sayln ""; indent keep_curpos;
		      prt_args_of_print (pos_percent+2) reste_args
		    end
		end
	    end

	      (* '%' not found in format string *)
	    with Not_found ->
	      begin
		try let pos_nl = find_slash_n current_pos fmt
		in
		begin
		  prt_print_call "Unquoted"; say ", \"";
		  say (String.sub fmt current_pos (pos_nl - current_pos));
		  say "\")";
                  (* faut-il encore recurser ? *)
		  if (must_recurse (pos_nl+2) liste_args)
		  then
		    begin
		      sayln " ;"; indent keep_curpos; prt_args_of_print (pos_nl+2) liste_args
		    end
		end
		with Not_found
		  ->
		    begin
		      prt_print_call "NoNL"; say ",";
		      if (current_pos < len)
		      then
			begin
			  say "\"";
			  say (String.sub fmt current_pos (len - current_pos));
			  say "\",";
			end;
		      prt_exprs_no_paren liste_args;
		      say ") "
		    end
	      end
	  end
    in prt_args_of_print 0 reste_args

  and prt_sprintf fmt args =
    let current_pos = ref 0
    and len = String.length fmt
    and c_est_parti = ref false
    in let tester_c_est_parti () =
      if !c_est_parti
      then say " . "
      else c_est_parti := true
    in
    let rec prt_sprintf_aux = function
      |	[] ->
	  begin
	    if (!current_pos < len)
	    then
	      begin
		tester_c_est_parti();
		say "\"";
		say (String.sub fmt !current_pos (len - !current_pos ));
		say "\"";
	      end
	    else ()
	  end
      |	x :: args ->
	  begin
	    try let pos_percent = (String.index_from fmt !current_pos '%')
	    in
	    begin
	      tester_c_est_parti();
	      say "\"";
	      say (String.sub fmt !current_pos (pos_percent - !current_pos));
	      say "\"";
	      if ( (pos_percent+1 < len) & (fmt.[pos_percent+1] = 's') )
	      then begin say " . "; prt_expr x end
	      else begin say ". expr2text ("; prt_expr x; say ")" end;
	      current_pos := 2 + pos_percent;
	      prt_sprintf_aux args
	    end
	    with Not_found ->
	      begin
		tester_c_est_parti();
		if (!current_pos < len)
		then
		  begin
		    say "\"";
		    say (String.sub fmt !current_pos (len - !current_pos ));
		    say "\"";
		    say " . "
		  end;
		say "expr2text ("; prt_expr x; say ")";
	      current_pos := len;
	      prt_sprintf_aux args
	      end
	  end
    in prt_sprintf_aux args

  (* traduction d'un statement *)
  (* l'argument boolen debut est faux
     ssi on veut commencer par l'impression d'un saut de ligne *)
  and prt_stat d debut = function
    | AssignStat (st) ->
	begin
	  if (not debut) then sayln "";
	  indent d;
	  begin
	    match st.lhs_of_assign with
	    | [ NameString(QuotedString(n)) ] when n.[0] = '&' ->
		begin
		  say "operator (";
		  say "\""; say n; say "\"";
		  say ", ";
		  prt_exprs_no_paren st.rhs;
		  sayln ","; indent (d+3);
		  say "Nary, 50";
		  say ")"
		end
	    | [ NameString(Ident("newvar")) ] -> say "newvar := newvar"
	    | [ NameString(Ident("newvars")) ] ->
		begin
		  say "newvars := (base_, lst_) -> map(lst_,elm_ -> genident ( base_ ))"
		end
	    | [ NameString(Ident(n)) ] ->
		begin
		  say (change_keywords n);
		  say " := ";
		  match st.rhs with
		  | [ NameExpr(NameString(QuotedString(m))) ] when m = n -> say "NIL"
		  | _ -> prt_exprs_no_paren st.rhs
		end
	    | [n] ->
		begin prt_name n; say " := "; prt_exprs_no_paren st.rhs end
	    | _ ->
		begin
		  say "["; prt_nameseq st.lhs_of_assign; say "]";
		  say " := ";
		  say "["; prt_exprs_no_paren st.rhs; say "]"
		end
	  end
	end
    | ExprStat (e) ->
	begin
	  if (not debut) then sayln "";
	  indent d; prt_expr e
	end
    | ReadStat (e) ->
	begin
	  if (not debut) then sayln "";
	  indent d; say "read ( "; prt_expr e; say " )"
	end
    | SaveStat (nl,fn)
      ->
	begin
	  if (not debut) then sayln "";
	  indent d; say "write ("; prt_name_string fn;
	  begin
	    match nl with
	    | [] -> ()
	    | _
	      ->
		let rec prt_nl = function
		  | [] -> ()
		  | n :: suite -> begin say ", "; prt_name n; prt_nl suite end
		in prt_nl nl
	  end;
	  say ")"
	end
    | IfThenElseStat (ite)
      ->
	let condition = (ite.cond : expr)
	and th_stmts = (ite.then_part : stat_comment_sep list)
	and alts = (ite.alternatives : alt list)
	and els_stmts = (ite.else_part : stat_comment_sep list)
	in
	begin
	  if (not debut) then sayln ""; indent d; say "if ";
	  prt_expr condition; sayln "";
	  indent (d+3); say " then";
	  prt_stmts (d+6) false th_stmts;
	  let rec prt_alternatives = function
	    | [] -> ()
	    | alt :: suite_alt
	      ->
		begin
		  sayln "";
		  indent (d+3); say "elif "; prt_expr alt.cond_of_alt; say " then ";
		  prt_stmts (d+6) false alt.stats_of_alt;
		  prt_alternatives suite_alt
		end
	  in prt_alternatives alts;
	  begin
	    match els_stmts with
	    | [] -> ()
	    | _
	      ->
		begin
		  sayln "";
		  indent (d+3); say "else";
		  prt_stmts (d+6) false els_stmts
		end
	  end ;
	  sayln ""; indent d; say "end_if"
	end
    | ForStat (forst)
	-> if is_a_while forst
	then
	  begin
	    if (not debut) then sayln "";
	    indent d; say "while";
	    begin
	      match forst.while_expr with
	      | Some e -> begin say " ("; prt_expr e; say ")" end
	      | _ -> say "(TRUE)"
	    end;
	    sayln " do";
	    prt_stmts (d+3) true forst.body;
	    sayln ""; indent d; say "end_while"
	  end
	else
	  begin
	    if (not debut) then sayln ""; indent d;
	    begin
	      match forst.loop_index with
	      |	Some n -> begin say "for "; prt_name n; say " " end
	      |	_ -> ()
	    end ;
	    begin
	      match forst.from_expr with
	      |	Some e -> begin say "from "; prt_expr e; say " " end
	      |	_ -> begin say "from 1 "; end
	    end ;
	    begin
	      match forst.to_expr with
	      |	Some e -> begin say "to "; prt_expr e; say " " end
	      |	_ -> ()
	    end ;
	    begin
	      match forst.by_expr with
	      |	Some e -> begin say "step "; prt_expr e; say " " end
	      |	_ -> ()
	    end ;
	    say "do ";
	    begin
	      match forst.while_expr with
	      |	Some e
		->
		  begin
		    sayln ""; indent (d+3);
		    say "if ( not "; prt_expr e; say " ) then break; end_if"
		  end
	      |	_ -> ()
	    end ;
	    prt_stmts (d+3) false forst.body;
	    sayln ""; indent d; say "end_for"
	  end
    | ForInStat (forst) ->
	let loop_variable =
	  match forst.loop_var with
	  | NameString (Ident (n)) ->
	      begin
		match n with
		| "assign" -> "__assign"   | "is" -> "__is"
		| "expr" -> "__expr"       | "hold" -> "__hold"
		| "left" -> "__left"       | "right" -> "__right"
		| "TRUE" -> "__TRUE"       | "FALSE" -> "__FALSE"
		| "null" -> "__null"
		| "PI" -> "__PI"           | "EULER" -> "__EULER"
		| _ -> n
	      end
	  | _ -> failwith "bad loop variable (for in)"
	in
	begin
          if (not debut) then sayln ""; indent d;
	  say "for ";
	  say loop_variable; say " in ";
	  prt_expr forst.in_expr;
	  say " do ";
	  begin
	    match forst.in_while with
	    | Some e ->
		begin
		  sayln ""; indent (d+3);
		  say "if ( not "; prt_expr e; say " ) then break; end_if"
		end
	    |	_ -> ()
	  end ;
	  prt_stmts (d+3) false forst.in_body;
	  sayln ""; indent d; say "end_for"
	end
    | StopStat -> begin if (not debut) then sayln ""; indent d; say "quit" end
    | EmptyStat -> ()
  and is_a_while forst =
    (match_none_name forst.loop_index)
      && (match_none_expr forst.from_expr)
      && (match_none_expr forst.by_expr)
      && (match_none_expr forst.to_expr)
  and prt_stmts d debut = function
    | [] -> ()
    | st :: suite ->
	begin
	  begin
	    match st.sep with
	    | SemiColon -> say " ;"
	    | Colon -> say " :"
	    | NoSep -> ()
	  end;
	  prt_comments (not debut) st.comments;
	  prt_stat d debut st.stat;
          prt_stmts d false suite
	end
  in
  begin
    match evaluation_level with
    | None -> ()
    | Some l ->
	begin
	  say "LEVEL := "; say_int l; sayln " :"
	end
  end;
  if (verbose) then
    cleartable();
  prt_stmts 0 true le_programme; sayln "";
  if (verbose) then
    begin
      if (not_empty()) then
	begin
	  Printf.printf "List of functions whose call has not been transformed:\n";
	  prttable()
	end;
      if (verbose) then cleartable()
    end
  ;;
